"""
Copyright (c) 2024, Auorui.
All rights reserved.

这里仅仅提供了一个激活函数的常用类型选择, 更多的激活函数手写实现你可以从‘pyzjr.nn.torchutils.function’中了解
"""
import torch
import torch.nn as nn

class ArgMax(nn.Module):
    def __init__(self, dim=None):
        super().__init__()
        self.dim = dim

    def forward(self, x):
        return torch.argmax(x, dim=self.dim)


class Clamp(nn.Module):
    def __init__(self, min=0, max=1):
        super().__init__()
        self.min, self.max = min, max

    def forward(self, x):
        return torch.clamp(x, self.min, self.max)


class Activation(nn.Module):
    def __init__(self, name, **params):
        super().__init__()
        if name is None or name == "identity":
            self.activation = nn.Identity(**params)
        elif name == "sigmoid":
            self.activation = nn.Sigmoid()
        elif name == "softmax":
            self.activation = nn.Softmax(dim=1, **params)
        elif name == "logsoftmax":
            self.activation = nn.LogSoftmax(dim=1, **params)
        elif name == "tanh":
            self.activation = nn.Tanh()
        elif name == "argmax":
            self.activation = ArgMax(**params)
        elif name == "clamp":
            self.activation = Clamp(**params)
        elif name == "relu":
            self.activation = nn.ReLU(inplace=True)
        elif callable(name):
            self.activation = name(**params)
        else:
            raise ValueError(
                f"Activation should be callable/sigmoid/softmax/logsoftmax/tanh/"
                f"argmax/clamp/relu/None; got {name}"
            )

    def forward(self, x):
        return self.activation(x)


import cv2
import torch
import numpy as np
from typing import Iterable
import torch.nn.functional as F

from pyzjr.utils.check import is_pil, is_tensor, is_numpy, is_list_or_tuple
from pyzjr.nn.strategy import preprocess_input


def to_numpy(x, dtype=None):
    if is_pil(x):
        return np.array(x, dtype=dtype)
    elif is_tensor(x):
        numpy_array = x.cpu().numpy()
        if dtype is not None:
            numpy_array = numpy_array.astype(dtype)
        return numpy_array
    elif is_numpy(x):
        if dtype is not None:
            return x.astype(dtype)
        return x
    elif isinstance(x, (Iterable, int, float)):
        return np.array(x, dtype=dtype)
    elif is_list_or_tuple(x):
        return np.array(x, dtype=dtype)
    else:
        raise ValueError("Unsupported type")

def to_tensor(x, dtype=None):
    if is_tensor(x):
        if dtype is not None:
            x = x.type(dtype)
        return x
    if is_numpy(x):
        x = torch.from_numpy(x)
        if dtype is not None:
            x = x.type(dtype)
        return x
    if is_list_or_tuple(x):
        x = np.array(x)
        x = torch.from_numpy(x)
        if dtype is not None:
            x = x.type(dtype)
        return x
    else:
        raise ValueError("Unsupported type")

def hwc2chw(img):
    """
    Conversion from 'HWC' to 'CHW' format.
    Example:
        hwc_image_numpy = np.random.rand(256, 256, 3)
        chw_image_numpy = hwc2chw(hwc_image_numpy)
        hwc_image_tensor = torch.rand(256, 256, 3)
        chw_image_tensor = hwc2chw(hwc_image_tensor)
    """
    if len(img.shape) == 3:
        if is_numpy(img):
            chw = np.transpose(img, axes=[2, 0, 1])
            return chw
        elif is_tensor(img):
            chw = img.permute(2, 0, 1).contiguous()
            return chw
        else:
            raise TypeError("The input data should be a NumPy array or "
                            "PyTorch tensor, but the provided type is: {}".format(type(img)))
    else:
        raise ValueError("The input data should be three-dimensional (height x width x channel), but the "
                         "provided number of dimensions is:{}".format(len(img.shape)))

def chw2hwc(img):
    """Conversion from 'CHW' to 'HWC' format."""
    if len(img.shape) == 3:
        if is_numpy(img):
            hwc = np.transpose(img, axes=[1, 2, 0])
            return hwc
        elif is_tensor(img):
            hwc = img.permute(1, 2, 0).contiguous()
            return hwc
        else:
            raise TypeError("The input data should be a NumPy array or "
                            "PyTorch tensor, but the provided type is: {}".format(type(img)))
    else:
        raise ValueError ("The input data should be three-dimensional (channel x height x width), but the "
                          "provided number of dimensions is: {}".format(len(img.shape)))

def to_bchw(tensor):
    """
    Convert to 'bchw' format
    Example:
        image_tensor = torch.rand(256, 256)
        bchw_image_tensor = to_bchw(image_tensor)
        print("Original shape:", image_tensor.shape)
        print("Converted shape:", bchw_image_tensor.shape)
    """
    if not isinstance(tensor, torch.Tensor):
        raise TypeError(f"Input type is not a Tensor. Got {type(tensor)}")

    if len(tensor.shape) < 2:
        raise ValueError(f"Input size must be a two, three or four dimensional tensor. Got {tensor.shape}")

    if len(tensor.shape) == 2:
        tensor = tensor.unsqueeze(0)

    if len(tensor.shape) == 3:
        tensor = tensor.unsqueeze(0)

    if len(tensor.shape) > 4:
        tensor = tensor.view(-1, tensor.shape[-3], tensor.shape[-2], tensor.shape[-1])

    return tensor

def image_to_bchw(image_data):
    """
    将图像加载为可输入网络的形状 b c h w, 且 b = 1
    """
    image_np = preprocess_input(np.array(image_data).astype(np.uint8))
    image_bchw = np.expand_dims(np.transpose(image_np, (2, 0, 1)), 0)
    return image_bchw

def bchw_to_image(image_bchw):
    """
    将网络输出转为图像类型, 且 b = 1
    """
    image_chw = image_bchw[0]  # chw
    image_hwc = image_chw.permute(1, 2, 0)
    image_np = F.softmax(image_hwc, dim=-1).cpu().numpy()
    image_np = np.clip(image_np * 255, 0, 255)
    image_np = image_np.astype(np.uint8)
    return image_np

def image_to_tensor(image, keepdim=True):
    """
    Convert numpy images to PyTorch 4d tensor images
    'keepdim' indicates whether to maintain the current dimension, otherwise it will be changed to type 4d
    Example:
        img = np.ones((3, 3))
        image_to_tensor(img).shape
    [1, 3, 3]
        img = np.ones((4, 4, 1))
        image_to_tensor(img).shape
    [1, 4, 4]
        img = np.ones((4, 4, 3))
        image_to_tensor(img, keepdim=False).shape
    [1, 3, 4, 4]
    """
    if is_numpy(image):
        if len(image.shape) > 4 or len(image.shape) < 2:
            raise ValueError("Input size must be a two, three or four dimensional array")
    input_shape = image.shape
    tensor = torch.from_numpy(image)

    if len(input_shape) == 2:
        # (H, W) -> (1, H, W)
        tensor = tensor.unsqueeze(0)
    elif len(input_shape) == 3:
        # (H, W, C) -> (C, H, W)
        tensor = tensor.permute(2, 0, 1)
    elif len(input_shape) == 4:
        # (B, H, W, C) -> (B, C, H, W)
        tensor = tensor.permute(0, 3, 1, 2)
        keepdim = True  # no need to unsqueeze
    else:
        raise ValueError(f"Cannot process image with shape {input_shape}")

    return tensor if keepdim else tensor.unsqueeze(0)

def tensor_to_image(tensor, keepdim = False):
    """Convert PyTorch tensor image to numpy image
    Returns:
        image of the form :math:`(H, W)`, :math:`(H, W, C)` or :math:`(B, H, W, C)`.
    Example:
        img = torch.ones(1, 3, 3)
        tensor_to_image(img).shape
    (3, 3)
        img = torch.ones(3, 4, 4)
        tensor_to_image(img).shape
    (4, 4, 3)
    """
    if not isinstance(tensor, torch.Tensor):
        raise TypeError(f"Input type is not a Tensor. Got {type(tensor)}")

    if len(tensor.shape) > 4 or len(tensor.shape) < 2:
        raise ValueError("Input size must be a two, three or four dimensional tensor")

    input_shape = tensor.shape
    image = tensor.cpu().detach().numpy()

    if len(input_shape) == 2:
        # (H, W) -> (H, W)
        pass
    elif len(input_shape) == 3:
        # (C, H, W) -> (H, W, C)
        if input_shape[0] == 1:
            # Grayscale for proper plt.imshow needs to be (H,W)
            image = image.squeeze()
        else:
            image = image.transpose(1, 2, 0)
    elif len(input_shape) == 4:
        # (B, C, H, W) -> (B, H, W, C)
        image = image.transpose(0, 2, 3, 1)
        if input_shape[0] == 1 and not keepdim:
            image = image.squeeze(0)
        if input_shape[1] == 1:
            image = image.squeeze(-1)
    else:
        raise ValueError(f"Cannot process tensor with shape {input_shape}")

    return image

def imagelist_to_tensor(imagelist):
    """Converts a list of numpy images to a PyTorch 4d tensor image.
    Args:
        images: list of images, each of the form :math:`(H, W, C)`.
        Image shapes must be consistent
    Returns:
        tensor of the form :math:`(B, C, H, W)`.
    Example:
        imgs = [np.ones((4, 4, 1)),
                np.zeros((4, 4, 1))]
        image_list_to_tensor(imgs).shape
    torch.Size([2, 1, 4, 4])
    """
    if len(imagelist[0].shape) != 3:
        raise ValueError("Input images must be three dimensional arrays")
    list_of_tensors = []
    for image in imagelist:
        list_of_tensors.append(image_to_tensor(image))
    return torch.stack(list_of_tensors)

def img2tensor(bgr, to_tensor=True):
    img = np.array(bgr).astype(np.uint8)
    rgb_img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
    chwrgb_img = np.moveaxis(rgb_img, -1, 0).astype(np.float32)
    return torch.from_numpy(chwrgb_img).type(torch.FloatTensor) if to_tensor else chwrgb_img

def label2tensor(mask, num_classes, sigmoid, totensor=False):
    """标签或掩码图像转换为 PyTorch 张量"""
    mask = np.array(mask)
    if num_classes > 2:
        if not sigmoid:
            long_mask = np.zeros((mask.shape[:2]), dtype=np.int64)
            if len(mask.shape) == 3:
                for c in range(mask.shape[2]):
                    long_mask[mask[..., c] > 0] = c
            else:
                long_mask[mask >= 127] = 1
                long_mask[mask == 0] = 0
            mask = long_mask
        else:
            mask = np.moveaxis(mask, -1, 0).astype(np.float32)
    else:
        mask[mask >= 127] = 1
        mask[mask == 0] = 0
    return torch.from_numpy(mask).long() if totensor else mask

if __name__=="__main__":
    hwc_image_numpy = np.random.rand(256, 256, 3)
    chw_image_numpy = hwc2chw(hwc_image_numpy)
    print("Original HWC shape:", hwc_image_numpy.shape)
    print("Converted CHW shape:", chw_image_numpy.shape)
    hwc_image_tensor = torch.rand(256, 256, 3)
    chw_image_tensor = hwc2chw(hwc_image_tensor)
    print("Original HWC shape:", hwc_image_tensor.shape)
    print("Converted CHW shape:", chw_image_tensor.shape)


import torch
import torch.nn as nn
import torch.nn.functional as F

class CrossEntropyLoss(nn.Module):
    """
    交叉熵损失（Cross Entropy Loss）用于多分类问题。
    用于测量预测输出和目标分布之间的交叉熵。依据公式实现。
    Args:
        input (torch.Tensor): The predicted output (logits).
        target (torch.Tensor): The target or ground truth (class labels).
        reduction (str, optional): Specifies the reduction to apply to the output.
            Options are 'none', 'mean', or 'sum'. Default is 'mean'.

    Examples::
        >>> criterion1 = nn.CrossEntropyLoss()
        >>> criterion2 = CrossEntropyLoss()
        >>> input_data = torch.randn((3, 5), requires_grad=True)
        >>> target_data = torch.randint(0, 5, (3,))
        >>> loss1 = criterion1(input_data, target_data)
        >>> loss2 = criterion2(input_data, target_data)
        >>> print("PyTorch CrossEntropyLoss:", loss1.item())
        >>> print("Custom CrossEntropyLoss:", loss2.item())

    Returns:
        torch.Tensor: The cross entropy loss between input and target.
    """
    def __init__(self, reduction='mean'):
        super(CrossEntropyLoss, self).__init__()
        self.reduction = reduction

    def forward(self, input, target):
        return nn.NLLLoss(reduction=self.reduction)(F.log_softmax(input, dim=1), target)

class LabelSmoothingCrossEntropy(nn.Module):
    """ NLL loss with label smoothing."""
    def __init__(self, smoothing=0.1):
        super(LabelSmoothingCrossEntropy, self).__init__()
        assert smoothing < 1.0
        self.smoothing = smoothing
        self.confidence = 1. - smoothing

    def forward(self, input, target):
        logprobs = F.log_softmax(input, dim=-1)
        nll_loss = -logprobs.gather(dim=-1, index=target.unsqueeze(1))
        nll_loss = nll_loss.squeeze(1)
        smooth_loss = -logprobs.mean(dim=-1)
        loss = self.confidence * nll_loss + self.smoothing * smooth_loss
        return loss.mean()


if __name__=="__main__":
    criterion1 = nn.CrossEntropyLoss()
    criterion2 = CrossEntropyLoss()
    criterion3 = LabelSmoothingCrossEntropy()
    input_data = torch.randn((3, 5), requires_grad=True)
    target_data = torch.randint(0, 5, (3,))
    loss1 = criterion1(input_data, target_data)
    loss2 = criterion2(input_data, target_data)
    loss3 = criterion3(input_data, target_data)
    print("PyTorch CrossEntropyLoss:", loss1.item())
    print("Custom CrossEntropyLoss:", loss2.item())
    print("Custom LabelSmoothingCrossEntropy:", loss3.item())


import random

def rand(a=0.0, b=1.0):
    """生成在指定范围内的随机浮点数,进行缩放和偏移来映射到[a, b)的范围。"""
    random_value = random.random()
    scaled_value = a + (b - a) * random_value
    return scaled_value

def randint(a=0, b=1):
    """生成一个指定范围内的随机整数，在 [a, b] 范围内的随机整数。"""
    return random.randint(a, b)

def randlist(size, a=0.0, b=1.0):
    """生成指定数量的随机浮点数列表，范围在 [a, b) 内。"""
    return [rand(a, b) for _ in range(size)]

def randintlist(size, a=0, b=1):
    """生成指定数量的随机整数列表，范围在 [a, b] 内。"""
    return [randint(a, b) for _ in range(size)]

def randbool():
    """生成一个随机的布尔值。"""
    return random.choice([True, False])

def randshuffle(lst):
    """随机打乱列表中的元素顺序。"""
    return random.shuffle(lst)

def randnormal(mu=0.0, sigma=1.0):
    """生成符合正态分布的随机数。"""
    return random.gauss(mu, sigma)

def randstring(length=8):
    """生成指定长度的随机字符串。"""
    characters = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789"
    return ''.join(random.choice(characters) for _ in range(length))

def randfluctuate(a=0.0, b=1.0, fluctuation=0.1):
    """生成指定区间内，带有波动的随机数。"""
    return rand(a - fluctuation, b + fluctuation)

def rand_choice_weighted(choices, weights):
    """根据给定的权重，从选项中随机选择一个。"""
    return random.choices(choices, weights=weights, k=1)[0]

def randstep(a=0.0, b=1.0, step=0.1):
    """生成指定步长范围内的随机浮点数。"""
    num_steps = int((b - a) // step)
    random_step = random.randint(0, num_steps)
    return a + random_step * step

if __name__ == "__main__":
    # 测试 rand() 函数
    print("rand() test:")
    for _ in range(5):
        print(rand(1.0, 5.0))

    # 测试 randint() 函数
    print("\nrandint() test:")
    for _ in range(5):
        print(randint(10, 20))

    # 测试 randlist() 函数
    print("\nrandlist() test:")
    print(randlist(5, 1.0, 10.0))

    # 测试 randintlist() 函数
    print("\nrandintlist() test:")
    print(randintlist(5, 1, 100))

    # 测试 randbool() 函数
    print("\nrandbool() test:")
    for _ in range(5):
        print(randbool())

    # 测试 randshuffle() 函数
    print("\nrandshuffle() test:")
    my_list = [1, 2, 3, 4, 5]
    randshuffle(my_list)
    print(my_list)

    # 测试 randnormal() 函数
    print("\nrandnormal() test:")
    for _ in range(5):
        print(randnormal(0.0, 1.0))

    # 测试 randstring() 函数
    print("\nrandstring() test:")
    for _ in range(5):
        print(randstring(8))

    # 测试 randfluctuate() 函数
    print("\nrandfluctuate() test:")
    for _ in range(5):
        print(randfluctuate(10.0, 20.0, 2.0))

    # 测试 rand_choice_weighted() 函数
    print("\nrand_choice_weighted() test:")
    choices = ['apple', 'banana', 'cherry']
    weights = [0.1, 0.3, 0.6]
    for _ in range(5):
        print(rand_choice_weighted(choices, weights))

    # 测试 randstep() 函数
    print("\nrandstep() test:")
    for _ in range(5):
        print(randstep(1.0, 10.0, 0.5))






"""
Original paper addresshttps: https://arxiv.org/pdf/1709.01507.pdf
Code originates from: https://github.com/moskomule/senet.pytorch/blob/master/senet/se_module.py
Blog records: https://blog.csdn.net/m0_62919535/article/details/135761713
Time: 2024-01-23
"""
import torch
from torch import nn

class SEAttention(nn.Module):
    def __init__(self, channel, reduction_ratio=16):
        super().__init__()
        self.avg_pool = nn.AdaptiveAvgPool2d(1)
        self.fc = nn.Sequential(
            nn.Linear(channel, channel // reduction_ratio, bias=False),
            nn.ReLU(inplace=True),
            nn.Linear(channel // reduction_ratio, channel, bias=False),
            nn.Sigmoid()
        )

    def forward(self, x):
        b, c, _, _ = x.size()
        y = self.avg_pool(x).view(b, c)
        y = self.fc(y).view(b, c, 1, 1)
        return x * y.expand_as(x)


class SqueezeExcitation(nn.Module):
    """
    This block implements the Squeeze-and-Excitation block from https://arxiv.org/abs/1709.01507 (see Fig. 1).
    Parameters ``activation``, and ``scale_activation`` correspond to ``delta`` and ``sigma`` in eq. 3.
    """
    def __init__(
            self,
            input_channels,
            squeeze_channels,
            activation=torch.nn.ReLU,
            scale_activation=torch.nn.Sigmoid):
        super().__init__()
        self.avgpool = nn.AdaptiveAvgPool2d(1)
        self.fc1 = nn.Conv2d(input_channels, squeeze_channels, 1)
        self.fc2 = nn.Conv2d(squeeze_channels, input_channels, 1)
        self.activation = activation()
        self.scale_activation = scale_activation()

    def _scale(self, input):
        scale = self.avgpool(input)
        scale = self.fc1(scale)
        scale = self.activation(scale)
        scale = self.fc2(scale)
        return self.scale_activation(scale)

    def forward(self, input):
        scale = self.avgpool(input)
        scale = self.fc1(scale)
        scale = self.activation(scale)
        scale = self.fc2(scale)
        scale = self.scale_activation(scale)
        return scale * input


class EffectiveSEAttention(nn.Module):
    """ 'Effective Squeeze-Excitation
    From `CenterMask : Real-Time Anchor-Free Instance Segmentation` - https://arxiv.org/abs/1911.06667
    code come from: https://github.com/youngwanLEE/CenterMask/blob/master/maskrcnn_benchmark/modeling/backbone/vovnet.py
    """
    def __init__(self, channel):
        super(EffectiveSEAttention, self).__init__()
        # self.avg_pool = nn.AdaptiveAvgPool2d(1)
        self.fc = nn.Conv2d(channel, channel, kernel_size=1,
                            padding=0)
        self.hsigmoid = nn.Hardsigmoid(inplace=True)

    def forward(self, x):
        input = x
        x = x.mean((2, 3), keepdim=True)   # x = self.avg_pool(x)
        x = self.fc(x)
        x = self.hsigmoid(x)
        return input * x


if __name__ == '__main__':
    input = torch.randn(1, 64, 64, 64)
    se = EffectiveSEAttention(channel=64)
    output = se(input)
    print(output.shape)



import torch
import torch.nn as nn

def _threshold(x, threshold=None):
    if threshold is not None:
        return (x > threshold).type(x.dtype)
    else:
        return x

def _take_channels(*xs, ignore_channels=None):
    if ignore_channels is None:
        return xs
    else:
        channels = [channel for channel in range(xs[0].shape[1]) if channel not in ignore_channels]
        xs = [torch.index_select(x, dim=1, index=torch.tensor(channels).to(x.device)) for x in xs]
        return xs

def iou(pr, gt, eps=1e-7, threshold=None, ignore_channels=None):
    """Calculate Intersection over Union between ground truth and prediction
    Args:
        pr (torch.Tensor): predicted tensor
        gt (torch.Tensor):  ground truth tensor
        eps (float): epsilon to avoid zero division
        threshold: threshold for outputs binarization
    Returns:
        float: IoU (Jaccard) score
    """
    pr = _threshold(pr, threshold=threshold)
    pr, gt = _take_channels(pr, gt, ignore_channels=ignore_channels)

    intersection = torch.sum(gt * pr)
    union = torch.sum(gt) + torch.sum(pr) - intersection + eps
    return (intersection + eps) / union


jaccard = iou


def f_score(pr, gt, beta=1, eps=1e-7, threshold=None, ignore_channels=None):
    """Calculate F-score between ground truth and prediction
    Args:
        pr (torch.Tensor): predicted tensor
        gt (torch.Tensor):  ground truth tensor
        beta (float): positive constant
        eps (float): epsilon to avoid zero division
        threshold: threshold for outputs binarization
    Returns:
        float: F score
    """
    pr = _threshold(pr, threshold=threshold)
    pr, gt = _take_channels(pr, gt, ignore_channels=ignore_channels)

    tp = torch.sum(gt * pr)
    fp = torch.sum(pr) - tp
    fn = torch.sum(gt) - tp

    score = ((1 + beta**2) * tp + eps) / ((1 + beta**2) * tp + beta**2 * fn + fp + eps)

    return score


def accuracy(pr, gt, threshold=0.5, ignore_channels=None):
    """Calculate accuracy score between ground truth and prediction
    Args:
        pr (torch.Tensor): predicted tensor
        gt (torch.Tensor):  ground truth tensor
        eps (float): epsilon to avoid zero division
        threshold: threshold for outputs binarization
    Returns:
        float: precision score
    """
    pr = _threshold(pr, threshold=threshold)
    pr, gt = _take_channels(pr, gt, ignore_channels=ignore_channels)

    tp = torch.sum(gt == pr, dtype=pr.dtype)
    score = tp / gt.view(-1).shape[0]
    return score


def precision(pr, gt, eps=1e-7, threshold=None, ignore_channels=None):
    """Calculate precision score between ground truth and prediction
    Args:
        pr (torch.Tensor): predicted tensor
        gt (torch.Tensor):  ground truth tensor
        eps (float): epsilon to avoid zero division
        threshold: threshold for outputs binarization
    Returns:
        float: precision score
    """
    pr = _threshold(pr, threshold=threshold)
    pr, gt = _take_channels(pr, gt, ignore_channels=ignore_channels)

    tp = torch.sum(gt * pr)
    fp = torch.sum(pr) - tp

    score = (tp + eps) / (tp + fp + eps)

    return score


def recall(pr, gt, eps=1e-7, threshold=None, ignore_channels=None):
    """Calculate Recall between ground truth and prediction
    Args:
        pr (torch.Tensor): A list of predicted elements
        gt (torch.Tensor):  A list of elements that are to be predicted
        eps (float): epsilon to avoid zero division
        threshold: threshold for outputs binarization
    Returns:
        float: recall score
    """
    pr = _threshold(pr, threshold=threshold)
    pr, gt = _take_channels(pr, gt, ignore_channels=ignore_channels)

    tp = torch.sum(gt * pr)
    fn = torch.sum(gt) - tp

    score = (tp + eps) / (tp + fn + eps)

    return score


class IoU_utils(nn.Module):
    def __init__(self, eps=1e-7, threshold=0.5, ignore_channels=None, **kwargs):
        super().__init__()
        self.eps = eps
        self.threshold = threshold
        self.ignore_channels = ignore_channels

    def forward(self, y_pr, y_gt):
        return iou(
            y_pr,
            y_gt,
            eps=self.eps,
            threshold=self.threshold,
            ignore_channels=self.ignore_channels,
        )


class Fscore_utils(nn.Module):
    def __init__(self, beta=1, eps=1e-7, threshold=0.5, ignore_channels=None, **kwargs):
        super().__init__()
        self.eps = eps
        self.beta = beta
        self.threshold = threshold
        self.ignore_channels = ignore_channels

    def forward(self, y_pr, y_gt):
        return f_score(
            y_pr,
            y_gt,
            eps=self.eps,
            beta=self.beta,
            threshold=self.threshold,
            ignore_channels=self.ignore_channels,
        )


class Accuracy_utils(nn.Module):
    def __init__(self, threshold=0.5, ignore_channels=None, **kwargs):
        super().__init__()
        self.threshold = threshold
        self.ignore_channels = ignore_channels

    def forward(self, y_pr, y_gt):
        return accuracy(
            y_pr,
            y_gt,
            threshold=self.threshold,
            ignore_channels=self.ignore_channels,
        )


class Recall_utils(nn.Module):
    def __init__(self, eps=1e-7, threshold=0.5, ignore_channels=None, **kwargs):
        super().__init__()
        self.eps = eps
        self.threshold = threshold
        self.ignore_channels = ignore_channels

    def forward(self, y_pr, y_gt):
        return recall(
            y_pr,
            y_gt,
            eps=self.eps,
            threshold=self.threshold,
            ignore_channels=self.ignore_channels,
        )


class Precision_utils(nn.Module):
    def __init__(self, eps=1e-7, threshold=0.5, ignore_channels=None, **kwargs):
        super().__init__()
        self.eps = eps
        self.threshold = threshold
        self.ignore_channels = ignore_channels

    def forward(self, y_pr, y_gt):
        return precision(
            y_pr,
            y_gt,
            eps=self.eps,
            threshold=self.threshold,
            ignore_channels=self.ignore_channels,
        )


import os
from pyzjr.data.base import timestr
from shutil import rmtree


def logdir(dir="logs", simple=True, prefix="", suffix=""):
    """
    Logging generator
    :param dir: Default "logs"
    :param simple: '%Y_%m_%d_%H_%M_%S' or '%A_%d_%B_%Y_%Hh_%Mm_%Ss'
    :param prefix: Add folder name before time string
    :param suffix: After adding the folder name to the time string
    :return:
    """
    time_str = timestr(simple)
    folder_names = [prefix, time_str, suffix]
    folder_names = [folder for folder in folder_names if folder]
    log_dir = os.path.join(dir, *folder_names)
    os.makedirs(log_dir, exist_ok=True)

    return log_dir

def multi_makedirs(*args):
    """
    为给定的多个路径创建目录, 如果路径不存在, 则创建它
    """
    for path in args:
        if not os.path.exists(path):
            os.makedirs(path)

def unique_makedirs(root_path, *paths, inc=False):
    """
    如果构建的输出目录路径不存在，它将创建该目录。
    如果 inc 为 True，它将检查是否已经存在相同的目录。如果存在，将在目录名称末尾附加一个数字后缀，直到找到一个不存在的目录。
    返回最终的输出目录路径。
    """
    outdir = os.path.join(root_path, *paths)
    if not os.path.exists(outdir):
        os.makedirs(outdir)
    elif inc:
        count = 1
        outdir_inc = outdir + '_' + str(count)
        while os.path.exists(outdir_inc):
            count = count + 1
            outdir_inc = outdir + '_' + str(count)
            assert count < 100
        outdir = outdir_inc
        os.makedirs(outdir)
    return outdir

def datatime_makedirs(root_path, *paths, use_simple_format=True):
    """
    创建含有独特时间的路径
    """
    time_path = timestr(use_simple_format)
    outdir = os.path.join(root_path, time_path, *paths)
    os.makedirs(outdir, exist_ok=True)
    return outdir

def loss_weights_dirs(root_path='./logs', use_simple_format=True):
    """训练代码常用路径, 保存loss信息, 以及保存最佳模型"""
    time_str = timestr(use_simple_format)
    time_dir = os.path.join(root_path, time_str)
    loss_log_dir = os.path.join(time_dir, 'loss')
    save_model_dir = os.path.join(time_dir, 'weights')
    multi_makedirs(loss_log_dir, save_model_dir)
    return loss_log_dir, save_model_dir, time_dir

def rm_makedirs(file_path: str):
    # 如果文件夹存在，则先删除原文件夹在重新创建
    if os.path.exists(file_path):
        rmtree(file_path)
    os.makedirs(file_path)

if __name__=="__main__":
    root_path = 'test_output'
    subdir_name = 'test_subdir'

    # outdir1 = unique_makedirs(root_path, subdir_name)
    # print(f"Created directory: {outdir1}")
    #
    # outdir2 = unique_makedirs(root_path, subdir_name, inc=True)
    # print(f"Created directory with inc: {outdir2}")
    #
    # outdir3 = unique_makedirs(root_path, subdir_name, inc=True)
    # print(f"Created directory with inc (second time): {outdir3}")
    # cleanup_test_dir(root_path)

    datatime_makedirs(root_path, "loss")
    datatime_makedirs(root_path, "weight")



from .io import (
    bgr_read, imwriter, display, url2image, stacked2image, StackedImagesV1, StackedImagesV2,
    imattributes, rgb_read, imshowplt, StackedpltV1, StackedpltV2, matplotlib_patch,
    VideoCap, Mp4toGif, FindColor, DetectImageColor, DetectVideoColor
)
from .plot import (
    AddText, PutMultiLineText, PutMultiLineCenteredText, PutBoxText,
    PutRectangleText, DrawPolygon, DrawCornerRectangle, OverlayPng, ConvertBbox
)
from .colorspace import (
    to_gray, rgb2bgr, bgr2rgb, to_hsv, hsv2rgb, hsv2bgr, pil2cv, cv2pil, create_palette
)
from .core import Timer, FPS, Runcodes, timing
from .printf import (
    ConsoleLogger, redirect_console, colorstr, colorfulstr, show_config, LoadingBar,
    printprocess, printlog, printcolor
)


"""
Copyright (c) 2023, Auorui.
All rights reserved.

部分手写实现,仅供学习参考,最好还是使用cv2
https://blog.csdn.net/m0_62919535/category_11936595.html
"""
import cv2
import numpy as np
from pyzjr.utils.check import is_odd, is_gray_image

def meanblur(img, ksize):
    """均值滤波 """
    blur_img = cv2.blur(img, ksize=ksize)
    return blur_img

def medianblur(img, ksize):
    """中值滤波"""
    if img.dtype == np.float32 and ksize not in {3, 5, 7, 9, 11}:
        raise ValueError(f"Invalid ksize value {ksize}.The available values are 3, 5, and 7")
    medblur_img = cv2.medianBlur(img, ksize=ksize)
    return medblur_img

def gaussianblur(img, ksize, sigma=None):
    """
    高斯模糊, 提供给不熟悉高斯模糊参数的用户, sigma根据ksize进行自动计算,具体可以参考下面
    https://docs.opencv.org/master/d4/d13/tutorial_py_filtering.html
    """
    if sigma is None:
        sigma = 0.3 * ((ksize - 1) * 0.5 - 1) + 0.8
    gaussianblur_img = cv2.GaussianBlur(img, ksize=(ksize, ksize), sigmaX=sigma)
    return gaussianblur_img

def bilateralblur(img, d=5, sigma_color=75, sigma_space=75):
    """双边滤波"""
    return cv2.bilateralFilter(img, d, sigma_color, sigma_space)

def medianfilter(image, ksize):
    assert is_odd(ksize), "Ksize must be an odd number"

    if is_gray_image(image):
        height, width = image.shape
        channels = 1
        image = image[:, :, np.newaxis]
    else:
        height, width, channels = image.shape

    half_window = ksize // 2
    padded_image = np.pad(image, ((half_window, half_window), (half_window, half_window), (0, 0)), mode='reflect')
    filtered_image = np.zeros_like(image)

    for c in range(channels):
        channel_data = padded_image[:, :, c]
        shape = (height, width, ksize, ksize)
        strides = (channel_data.strides[0], channel_data.strides[1], channel_data.strides[0], channel_data.strides[1])
        windows = np.lib.stride_tricks.as_strided(channel_data, shape=shape, strides=strides)
        filtered_image[:, :, c] = np.median(windows, axis=(2, 3))
    if channels == 1:
        filtered_image = filtered_image[:, :, 0]
    return filtered_image

def meanfilter(image, ksize):
    assert is_odd(ksize), "Ksize must be an odd number"

    if is_gray_image(image):
        height, width = image.shape
        channels = 1
        image = image[:, :, np.newaxis]
    else:
        height, width, channels = image.shape

    half_window = ksize // 2
    padded_image = np.pad(image, ((half_window, half_window), (half_window, half_window), (0, 0)), mode='reflect')
    filtered_image = np.zeros_like(image)

    for c in range(channels):
        channel_data = padded_image[:, :, c]
        shape = (height, width, ksize, ksize)
        strides = (channel_data.strides[0], channel_data.strides[1], channel_data.strides[0], channel_data.strides[1])
        windows = np.lib.stride_tricks.as_strided(channel_data, shape=shape, strides=strides)
        filtered_image[:, :, c] = np.mean(windows, axis=(2, 3))

    if channels == 1:
        filtered_image = filtered_image[:, :, 0]

    return filtered_image

def gaussian_kernel(size, sigma):
    center = size // 2
    y, x = np.ogrid[-center:center+1, -center:center+1]
    kernel = np.exp(-(x**2 + y**2) / (2 * sigma**2))
    kernel /= kernel.sum()
    return kernel

def gaussianfilter(image, ksize, sigma=None):
    """
    sigma 可参考此处
    https://docs.opencv.org/master/d4/d13/tutorial_py_filtering.html
    """
    assert is_odd(ksize), "Ksize must be an odd number"
    if sigma is None:
        sigma = 0.3 * ((ksize - 1) * 0.5 - 1) + 0.8

    if is_gray_image(image):
        height, width = image.shape
        channels = 1
        image = image[:, :, np.newaxis]
    else:
        height, width, channels = image.shape

    half_window = ksize // 2
    padded_image = np.pad(image, ((half_window, half_window), (half_window, half_window), (0, 0)), mode='reflect')
    filtered_image = np.zeros_like(image)
    kernel = gaussian_kernel(ksize, sigma)

    for c in range(channels):
        channel_data = padded_image[:, :, c]
        shape = (height, width, ksize, ksize)
        strides = (channel_data.strides[0], channel_data.strides[1], channel_data.strides[0], channel_data.strides[1])
        windows = np.lib.stride_tricks.as_strided(channel_data, shape=shape, strides=strides)
        filtered_image[:, :, c] = np.tensordot(windows, kernel, axes=((2, 3), (0, 1)))

    if channels == 1:
        filtered_image = filtered_image[:, :, 0]

    return filtered_image

def bilateralfilter(image, d=5, sigma_color=75, sigma_space=75):
    def gaussian(x, sigma):
        return np.exp(-(x**2) / (2 * sigma**2))

    assert is_odd(d), "Diameter must be an odd number"

    if is_gray_image(image):
        height, width = image.shape
        channels = 1
        image = image[:, :, np.newaxis]
    else:
        height, width, channels = image.shape

    half_d = d // 2
    padded_image = np.pad(image, ((half_d, half_d), (half_d, half_d), (0, 0)), mode='reflect')
    filtered_image = np.zeros_like(image)

    y, x = np.ogrid[-half_d:half_d+1, -half_d:half_d+1]
    spatial_weights = gaussian(np.sqrt(x**2 + y**2), sigma_space)

    for c in range(channels):
        channel_data = padded_image[:, :, c]

        for i in range(height):
            for j in range(width):
                neighborhood = channel_data[i:i+d, j:j+d]
                color_weights = gaussian(neighborhood - channel_data[i + half_d, j + half_d], sigma_color)
                weights = spatial_weights * color_weights
                weights /= np.sum(weights)
                filtered_pixel = np.sum(weights * neighborhood)
                filtered_image[i, j, c] = filtered_pixel

    if channels == 1:
        filtered_image = filtered_image[:, :, 0]

    return filtered_image

if __name__=="__main__":
    import pyzjr
    image = pyzjr.bgr_read("dog.png", is_gray=False)
    with pyzjr.Runcodes():
        image = bilateralfilter(image)
    pyzjr.display(image)


import torch
import sys
from pyzjr.nn.models.backbone import *
from pyzjr.nn.devices import load_owned_device

def get_clsnetwork(name, num_classes=None, weights='', device=load_owned_device()):
    if name == 'LeNet':
        net = LeNet(num_classes=10)
    else:
        network_names = [
            'ZFNet', 'AlexNet', 'vgg11_bn', 'vgg13_bn', 'vgg16_bn', 'vgg19_bn',
            'googlenet', 'resnet18', 'resnet34', 'resnet50', 'resnet101', 'resnet152',
            'conv2former_n', 'conv2former_t', 'conv2former_s', 'conv2former_l', 'conv2former_b',
            'se_resnet18', 'se_resnet34', 'se_resnet50', 'se_resnet101', 'se_resnet152',
            'squeezenet1_0', 'squeezenet1_1', 'darknet19', 'darknet53',
            'MobileNetV1', 'MobileNetV2', 'MobileNetV3_Large', 'MobileNetV3_Small',
            'densenet121', 'densenet161', 'densenet169', 'densenet201',
            'mnasnet0_5', 'mnasnet0_75', 'mnasnet1_0', 'mnasnet1_3', 'shufflenet_v1_g1',
            'shufflenet_v1_g2', 'shufflenet_v1_g3', 'shufflenet_v1_g4', 'shufflenet_v1_g8',
            'shufflenet_v2_x0_5', 'shufflenet_v2_x1_0', 'shufflenet_v2_x1_5',
            'shufflenet_v2_x2_0', 'Xception', 'drn_c_26', 'drn_c_42', 'drn_c_58', 'drn_d_22',
            'drn_d_24', 'drn_d_38', 'drn_d_40', 'drn_d_54', 'drn_d_56', 'drn_d_105', 'drn_d_107',
            'ghostnetv1', 'ghostnetv2', 'g_ghost_regnetx_002', 'g_ghost_regnetx_004',
            'g_ghost_regnetx_006', 'g_ghost_regnetx_008', 'g_ghost_regnetx_016',
            'g_ghost_regnetx_032', 'g_ghost_regnetx_040', 'g_ghost_regnetx_064',
            'g_ghost_regnetx_080', 'g_ghost_regnetx_120', 'g_ghost_regnetx_160',
            'g_ghost_regnetx_320', 'regnetx_002', 'regnetx_004', 'regnetx_006', 'regnetx_008',
            'regnetx_016', 'regnetx_032', 'regnetx_040', 'regnetx_064', 'regnetx_080',
            'regnetx_120', 'regnetx_160', 'regnetx_320', 'fasternet_t0', 'fasternet_t1',
            'fasternet_t2', 'fasternet_s', 'fasternet_m', 'fasternet_l',  'vit_b_16', 'vit_b_32',
            'vit_l_16', 'vit_l_32', 'vit_h_14', 'swin_t', 'swin_s', 'swin_b', 'swin_l'
        ]

        if name not in network_names:
            print('the network name you have entered is not supported yet')
            sys.exit()

        network_class = getattr(sys.modules[__name__], name)
        net = network_class(num_classes=num_classes)

    net.to(device)
    if weights != '':
        print('Load weights {}.'.format(weights))
        model_dict = net.state_dict()
        pretrained_dict = torch.load(weights, map_location=device)

        matched_keys = set(model_dict.keys()) & set(pretrained_dict.keys())
        mismatched_keys = set(model_dict.keys()) - set(pretrained_dict.keys())

        for key in matched_keys:
            if model_dict[key].shape == pretrained_dict[key].shape:
                model_dict[key] = pretrained_dict[key]
            else:
                mismatched_keys.add(key)

        if mismatched_keys:
            print("The following keys have mismatched shapes or are not present in the model:")
            for key in mismatched_keys:
                print(f"- {key}")
        else:
            print("All weights were successfully loaded.")
            net.load_state_dict(model_dict)
    else:
        print("\033[31mNo weights, training will start from scratch")

    return net

if __name__=="__main__":
    import torchsummary
    device = 'cuda' if torch.cuda.is_available() else 'cpu'
    input = torch.ones(2, 3, 224, 224).to(device)
    net = get_clsnetwork(name='vit_l_16', num_classes=4)
    net = net.to(device)
    out = net(input)
    print(out)
    print(out.shape)
    torchsummary.summary(net, input_size=(3, 224, 224))


"""
Copyright (c) 2024, Auorui.
All rights reserved.
time 2024-01-25
"""
import torch
import torch.nn as nn
import torch.nn.functional as F

__all__ = ["L1Loss", "L2Loss", "BCELoss", "Joint2loss", "MCCLoss"]

class L1Loss(nn.Module):
    """
    L1损失，也称为平均绝对误差（MAE），测量预测输出中的每个元素与目标或地面实况中的相应元素之间的平均绝对差。
    在数学上，它表示为预测值和目标值之间差异的绝对值的平均值。与L2损耗相比，L1损耗对异常值不那么敏感。依据公式实现。
    Args:
        input (torch.Tensor): The predicted output.
        target (torch.Tensor): The target or ground truth.
        reduction (str, optional): Specifies the reduction to apply to the output.
            Options are 'none', 'mean', or 'sum'. Default is 'mean'.
    Examples::
        >>> criterion1 = nn.L1Loss()
        >>> criterion2 = L1Loss()
        >>> input_data=torch.Tensor([2, 3, 4, 5])
        >>> target_data=torch.Tensor([4, 5, 6, 7])
        >>> loss1 = criterion1(input_data, target_data)  # tensor(2.)
        >>> loss2 = criterion2(input_data, target_data)  # tensor(2.)
    Returns:
        torch.Tensor: The L1 loss between input and target.
    """
    def __init__(self):
        super(L1Loss, self).__init__()

    def forward(self, input, target):
        loss = torch.mean(torch.abs(input - target))
        return loss

class L2Loss(nn.Module):
    """
    L2损失，也称为均方误差（MSE），测量预测输出中的每个元素与目标或地面实况中的相应元素之间的平均平方差。
    在数学上，它表示为预测值和目标值之间差异的平方的平均值。相比于L1损耗，L2损耗对异常值更敏感。依据公式实现。
    在torch当中是MSELoss
    Args:
        input (torch.Tensor): The predicted output.
        target (torch.Tensor): The target or ground truth.
    Examples::
        >>> criterion1 = nn.MSELoss()
        >>> criterion2 = L2Loss()
        >>> input_data=torch.Tensor([2, 3, 4, 5])
        >>> target_data=torch.Tensor([4, 5, 6, 7])
        >>> loss1 = criterion1(input_data, target_data)  # tensor(4.)
        >>> loss2 = criterion2(input_data, target_data)  # tensor(4.)

    Returns:
        torch.Tensor: The L2 loss between input and target.
    """
    def __init__(self, eps=1e-3):
        super(L2Loss, self).__init__()
        self.esp = eps

    def forward(self, input, target):
        loss = torch.mean(torch.pow(input - target, 2) + self.esp)
        return loss

class BCELoss(nn.Module):
    """
    二元交叉熵损失（Binary Cross Entropy Loss），也称为对数损失。
    用于测量预测输出中的每个元素与目标或地面实况中的相应元素之间的对数概率差异。依据公式实现。
    Args:
        input (torch.Tensor): The predicted output.Map to (0,1) through sigmoid function.
        target (torch.Tensor): The target or ground truth.
        reduction (str, optional): Specifies the reduction to apply to the output.
            Options are 'none', 'mean', or 'sum'. Default is 'mean'.

    Examples::
        >>> criterion1 = nn.BCELoss()
        >>> criterion2 = BCELoss()
        >>> input_data = torch.randn((5,))
        >>> target_data = torch.randint(0, 2, (5,), dtype=torch.float32)
        >>> loss1 = criterion1(torch.sigmoid(input_data), target_data)
        >>> loss2 = criterion2(input_data, target_data)
        >>> print("PyTorch BCELoss:", loss1.item())
        >>> print("Custom BCELoss:", loss2.item())

    Returns:
        torch.Tensor: The binary cross entropy loss between input and target.
    """
    def __init__(self, ignore_index=None, reduction='mean'):
        super(BCELoss, self).__init__()
        self.reduction = reduction
        self.ignore_index = ignore_index

    def forward(self, input, target):
        if self.ignore_index:
            input = input[:,:self.ignore_index,...]
            target = target[:, :self.ignore_index, ...]
        input = torch.sigmoid(input)
        loss = - (target * torch.log(input) + (1 - target) * torch.log(1 - input))
        if self.reduction == 'mean':
            return loss.mean()
        elif self.reduction == 'sum':
            return loss.sum()
        elif self.reduction == 'none':
            return loss
        else:
            raise ValueError(f'Unsupported reduction: {self.reduction}, available options are ["mean", "sum", "none"].')


class Joint2loss(nn.Module):
    """
    联合损失函数, 传入两个损失函数
        >>> criterion1 = FocalLoss()
        >>> criterion2 = DiceLoss()
        >>> joint_loss = Joint2loss(criterion1, criterion2, alpha=0.7, reduction='mean')
        >>> input_tensor = torch.rand((1,2,16,16), dtype=torch.float32)
        >>> target_tensor = torch.randn((1,2,16,16), dtype=torch.float32)
        >>> loss = joint_loss(input_tensor, target_tensor)
        >>> print("Joint Loss:", loss.item())
    """
    def __init__(self, *args, alpha, beta=None, ignore_index=None, reduction='mean'):
        super(Joint2loss, self).__init__()
        self.reduction = reduction
        self.ignore_index = ignore_index
        self.alpha = alpha
        self.beta = beta if beta is not None else 1-self.alpha
        self.criterion_1, self.criterion_2 = args

    def forward(self, input, target):
        if self.ignore_index:
            input = input[:, :self.ignore_index,...]
            target = target[:, :self.ignore_index, ...]
        loss_1 = self.criterion_1(input, target)
        loss_2 = self.criterion_2(input, target)
        loss = self.alpha * loss_1 + self.beta * loss_2
        if self.reduction == 'mean':
            return loss.mean()
        elif self.reduction == 'sum':
            return loss.sum()
        elif self.reduction == 'none':
            return loss
        else:
            raise ValueError(f'Unsupported reduction: {self.reduction}, available options are ["mean", "sum", "none"].')

class MCCLoss(nn.Module):
    def __init__(self, eps: float = 1e-5, threshold: float =.5):
        """Compute Matthews Correlation Coefficient Loss for image segmentation task.
        It only supports binary mode.

        Args:
            eps (float): Small epsilon to handle situations where all the samples in the dataset belong to one class

        Reference:
            https://github.com/kakumarabhishek/MCC-Loss
        """
        super().__init__()
        self.eps = eps
        self.threshold = threshold

    def forward(self, y_pred: torch.Tensor, y_true: torch.Tensor) -> torch.Tensor:
        """Compute MCC loss
        MCC = (TP.TN - FP.FN) / sqrt((TP+FP) . (TP+FN) . (TN+FP) . (TN+FN))
        Args:
            y_pred (torch.Tensor): model prediction of shape (N, H, W) or (N, 1, H, W)
            y_true (torch.Tensor): ground truth labels of shape (N, H, W) or (N, 1, H, W)

        Returns:
            torch.Tensor: loss value (1 - mcc)
        """
        y_pred = (y_pred >= self.threshold).float()
        bs = y_true.shape[0]

        y_true = y_true.view(bs, 1, -1)
        y_pred = y_pred.view(bs, 1, -1)

        tp = torch.sum(torch.mul(y_pred, y_true)) + self.eps
        tn = torch.sum(torch.mul((1 - y_pred), (1 - y_true))) + self.eps
        fp = torch.sum(torch.mul(y_pred, (1 - y_true))) + self.eps
        fn = torch.sum(torch.mul((1 - y_pred), y_true)) + self.eps

        numerator = torch.mul(tp, tn) - torch.mul(fp, fn)
        denominator = torch.sqrt(torch.add(tp, fp) * torch.add(tp, fn) * torch.add(tn, fp) * torch.add(tn, fn))

        mcc = torch.div(numerator.sum(), denominator.sum())
        loss = 1.0 - mcc

        return loss


class PSNRLoss(nn.Module):
    def __init__(self, loss_weight=1.0, reduction='mean', max_val=1):
        super(PSNRLoss, self).__init__()
        assert reduction == 'mean'
        self.loss_weight = loss_weight
        self.max_val = max_val

    def forward(self, pred, target):
        assert len(pred.size()) == 4
        mse = torch.mean((pred - target) ** 2, dim=(1, 2, 3))
        psnr = 10 * torch.log10((self.max_val ** 2) / (mse + 1e-8))
        return -self.loss_weight * torch.mean(psnr)

if __name__=="__main__":
    batch_size = 2
    height, width = 4, 4
    y_true = torch.randint(0, 2, (batch_size, 1, height, width)).float()
    y_pred = torch.rand((batch_size, 1, height, width))

    mcc_loss = MCCLoss()
    loss = mcc_loss(y_pred, y_true)
    print(f'MCC Loss: {loss.item()}')

    y_pred = torch.randn(1, 1, 2, 2)
    y_true = torch.randn(1, 1, 2, 2)
    psnr_loss_fn = PSNRLoss(loss_weight=1.0)
    loss = psnr_loss_fn(y_pred, y_true)
    print("PSNR Loss:", loss.item())


import cv2
import numpy as np
import pyzjr.Z as Z

def DrawPolygon(img, points, color):
    """
    Draw a polygon onto an image using the given points and fill color.
    Parameters:
    -----------
    img : numpy.ndarray
        The image (as a NumPy array) on which the rectangle and corners will be drawn.
    points : list of tuples or numpy.ndarray
            List of vertex coordinates for polygons. Each coordinate should be a tuple (x, y) containing two integers,
            Alternatively, a NumPy array with the shape (n, 1, 2) can be passed, where n is the number of vertices of the polygon.
    color : tuple
        Fill the color of the polygon in the format of (B, G, R), with each value ranging from 0 to 255 (inclusive)
    Code Example:
    ------------
    >>> image = np.ones((400, 400, 3), dtype=np.uint8)
    >>> polygon_points = np.array([[100, 100], [200, 50], [300, 150], [250, 300], [150, 250]])
    >>> result_image = DrawPolygon(image, polygon_points,  (255, 0, 0))
    """
    points = np.array([points])
    points = points.astype(np.int32)
    img = cv2.fillPoly(img, points, color)
    return img

def DrawCornerRectangle(img, bbox, length=30, lthickness=5, rthickness=1,
                        bboxcolor=Z.purple, cornercolor=Z.green):
    """
    Draws a rectangle with highlighted corners on an image. The rectangle is defined by a bounding box,
    and the corners are emphasized with lines of a specified length and thickness.

    Parameters:
    -----------
    img : numpy.ndarray
        The image (as a NumPy array) on which the rectangle and corners will be drawn.
    bbox : tuple
        The bounding box coordinates in the format `(x, y, x1, y1)`, where `(x, y)` is the top-left corner
        and `(x1, y1)` is the bottom-right corner.
    length : int, optional
        The length of the corner lines (default is 30 pixels).
    lthickness : int, optional
        The thickness of the corner lines (default is 5).
    rthickness : int, optional
        The thickness of the rectangle lines (default is 1).
    bboxcolor : tuple, optional
        The color of the rectangle in BGR format (e.g., (128, 0, 128) for purple). Default is `Z.purple`.
    cornercolor : tuple, optional
        The color of the corner lines in BGR format (e.g., (0, 255, 0) for green). Default is `Z.green`.

    Returns:
    --------
    numpy.ndarray
        The modified image with the rectangle and corners drawn.

    Usage Example:
    --------------
    >>> img = np.zeros((300, 500, 3), dtype=np.uint8)
    >>> bbox = (50, 50, 200, 150)
    >>> img = DrawCornerRectangle(img, bbox, length=30, lthickness=5, rthickness=1,
    ...                          bboxcolor=(128, 0, 128), cornercolor=(0, 255, 0))
    """
    x, y, x1, y1 = bbox
    cv2.rectangle(img, (x, y), (x1, y1), bboxcolor, rthickness)
    # Top-left corner (x, y)
    cv2.line(img, (x, y), (x + length, y), cornercolor, lthickness)
    cv2.line(img, (x, y), (x, y + length), cornercolor, lthickness)
    # Top-right corner (x1, y)
    cv2.line(img, (x1, y), (x1 - length, y), cornercolor, lthickness)
    cv2.line(img, (x1, y), (x1, y + length), cornercolor, lthickness)
    # Bottom-left corner (x, y1)
    cv2.line(img, (x, y1), (x + length, y1), cornercolor, lthickness)
    cv2.line(img, (x, y1), (x, y1 - length), cornercolor, lthickness)
    # Bottom-right corner (x1, y1)
    cv2.line(img, (x1, y1), (x1 - length, y1), cornercolor, lthickness)
    cv2.line(img, (x1, y1), (x1, y1 - length), cornercolor, lthickness)

    return img


if __name__=="__main__":
    imagePath = r"E:\PythonProject\pyzjrPyPi\pyzjr\test.png"
    image = np.ones((400, 400, 3), dtype=np.uint8)
    polygon_points = np.array([[100, 100], [200, 50], [300, 150], [250, 300], [150, 250]])
    result_image = DrawPolygon(image, polygon_points, (255, 0, 0))
    img = cv2.imread(imagePath)
    # img = np.zeros((300, 500, 3), dtype=np.uint8)
    bbox = (50, 50, 200, 150)
    img = DrawCornerRectangle(img, bbox, length=30, lthickness=5, rthickness=1,
                            bboxcolor=(128, 0, 128), cornercolor=(0, 255, 0))
    cv2.imshow("show", img)
    cv2.imshow("show2", result_image)
    cv2.waitKey(0)


"""
Copyright (c) 2025, Auorui.
All rights reserved.

对图像的标签进行处理, 包括灰度、二值化、骨架提取、轮廓提取、距离变换等等
"""
from .predeal import (clip, unique, create_rectmask, binarization, approximate, ceilfloor,
                      bool2mask, up_low, inpaint_defect, convert_mask, cvt8png, auto_canny,
                      mask_foreground_move)
from .skeleton_extraction import medial_axis_mask, skeletonizes, thinning, read_skeleton
from .distance_transform import cv_distance, chamfer, fast_marching
from .statistical_pixels import (count_nonzero, count_white, count_zero, unique, incircleV1,
                                incircleV2, outcircle)
from .contour import (SearchOutline, check_points_in_contour, getContours, foreground_contour_length,
                      sort_contours, label_contour, gradientOutline, drawOutline)
from .steger import Steger



import cv2
import numpy as np
from PIL import Image
from pyzjr.utils.check import is_bool, is_gray_image

def clip(image, maxval=255, dtype=np.uint8):
    """
    Truncate the pixel values of the image to the specified range and perform data type conversion
    """
    np_image = np.array(image)
    return np.clip(np_image, 0, maxval).astype(dtype)

def unique(image):
    """Returns a list of unique pixel values in the input image"""
    np_image = np.array(image).astype(np.uint8)
    return np.unique(np_image)

def create_rectmask(image, up_left_coord=None, lower_right_coord=None, bboxes=None):
    """
    Create a rectangular mask
    :param image: Original image.
    :param up_left_coord: [x1, y1]
    :param lower_right_coord: [x2, y2]
    :param bboxes: [x1, y1, x2, y2]
    """
    image = np.array(image).astype(np.uint8)
    h, w = image.shape[:2]
    mask = np.zeros((h, w, 3), dtype=np.uint8)
    if bboxes is None:
        x1, y1 = up_left_coord
        x2, y2 = lower_right_coord
    else:
        x1, y1, x2, y2 = bboxes
    mask[y1: y2, x1: x2] = 255

    if mask.ndim == 3 and mask.shape[-1] > 1:
        mask = mask[:, :, 0]
    masked_image = cv2.bitwise_and(image, image, mask=mask)
    return mask, masked_image


def binarization(image, min_value=127, max_value=255):
    """
    Convert image to binary image
    :param image: Original image. --- (BGR format)
    :param min_value: Minimum threshold, default to 127
    :param max_value: Maximum threshold, default to 225
    """
    np_image = np.array(image).astype(np.uint8)
    gray_image = cv2.cvtColor(np_image, cv2.COLOR_BGR2GRAY)
    _, binary_image = cv2.threshold(gray_image, min_value, max_value, cv2.THRESH_BINARY)
    return binary_image

def approximate(image, std=127.5, dtype=np.uint8):
    """
    Convert a single channel image into a binary image.
    """
    if not is_gray_image(image):
        image = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
    image[image > std] = 255
    image[image < std] = 0
    image = image.astype(dtype)
    return image

def ceilfloor(image, dtype=np.uint8):
    """
    The pixel value of the input image is limited between the maximum value of 255 and the minimum value of 0
    """
    if not is_gray_image(image):
        image = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
    image[image > 255] = 255
    image[image < 0] = 0
    image = image.astype(dtype)
    return image

def up_low(image, lower, upper, dtype=np.uint8):
    """
    Create a binarized mask based on the given color range (lower and upper limits)
    :param lower: Lower limit of color range, (B, G, R)
    :param upper: Upper limit of color range, (B, G, R)
    """
    np_image = np.array(image).astype(dtype)
    lower = np.array(lower, dtype=dtype)
    upper = np.array(upper, dtype=dtype)
    _mask = cv2.inRange(np_image, lower, upper)
    return _mask

def bool2mask(matrix, value=255):
    """
    Convert Boolean matrix to 8-bit unsigned integer mask image
    """
    if is_bool(matrix):
        result_int = matrix.astype(int)
        _mask = result_int * value
        return _mask.astype(np.uint8)
    else:
        raise ValueError("Input matrix must be of bool dtype.")

def inpaint_defect(image, mask, radius=10, flags=1):
    """
    Inpaint defect image
    :param image: defect image
    :param mask: Mask of defective parts
    :param radius: Inpaint radius
    :param flags: TELEA-1; NS-0
    """
    dst = cv2.inpaint(image, mask, radius, flags)
    return dst


def convert_mask(img_path, lower, upper, convert=(255, 255, 255)):
    """
    去除标签中的杂色,并对其进行颜色的转换

    :param img_path: 输入图像的文件路径
    :param lower: 颜色范围的下限值，作为一个包含(B, G, R)值的元组
    :param upper: 颜色范围的上限值，作为一个包含(B, G, R)值的元组
    :param convert: 需要转换为的颜色，作为一个包含(B, G, R)值的元组
    :return: 处理后的图像
    """
    image = cv2.imread(img_path)
    mask = up_low(image, lower, upper)
    image[mask > 0] = convert

    return image

def cvt8png(pngpath, bit_depth=False, target=(255, 255, 255), convert=(128, 0, 0)):
    """
    Voc: RGB png彩图转换
    :param bit_depth: 默认转为8位,需要使用out_png.save可以正确保存
    True:                      False:
        plt.imshow(img)             cv2.imshow("img",img)
        plt.axis('off')             cv2.waitKey(0)
        plt.show()
    :param pngpath: 为保证是按照cv2的方式读入, 所以传入路径即可
    :param target: 目标颜色, RGB方式
    :param convert: 转换颜色, 同RGB格式
    :return:
    """
    png = cv2.imread(pngpath)
    png = cv2.cvtColor(png, cv2.COLOR_BGR2RGB)
    h, w = png.shape[:2]

    mask = np.all(png == target, axis=-1)
    out_png = np.zeros((h, w, 3), dtype=np.uint8)
    out_png[mask] = convert
    out_png[~mask] = png[~mask]

    if bit_depth:
        out_png_pil = Image.fromarray(out_png)
        out_png_pil = out_png_pil.convert("P", palette=Image.Palette.ADAPTIVE, colors=256)
        return out_png_pil
    else:
        out_png_cv = cv2.cvtColor(out_png, cv2.COLOR_RGB2BGR)
        return out_png_cv


def mask_foreground_move(image, H_pixels=None, V_pixels=None):
    if H_pixels is None:
        H_pixels = 0
    if V_pixels is None:
        V_pixels = 0
    h, w = image.shape[:2]
    black = np.zeros_like(image)

    start_row, end_row = max(0, V_pixels), min(h, h + V_pixels)
    start_col, end_col = max(0, H_pixels), min(w, w + H_pixels)

    fill_start_row, fill_end_row = max(0, -V_pixels), min(h, h - V_pixels)
    fill_start_col, fill_end_col = max(0, -H_pixels), min(w, w - H_pixels)

    black[start_row:end_row, start_col:end_col] = image[fill_start_row:fill_end_row, fill_start_col:fill_end_col]

    return black

def auto_canny(image, sigma=0.33):
    """
    Automatically detect edges in the input image using the Canny edge detection algorithm.

    The lower and upper thresholds for Canny edge detection are determined based on the median
    pixel intensity of the image and a specified sigma value.

    Args:
        image (numpy.ndarray): Input image as a NumPy array (grayscale).
        sigma (float, optional): A scaling factor to determine the lower and upper thresholds.
            Default is 0.33.

    Returns:
        numpy.ndarray: A binary image with edges detected.

    """
    v = np.median(image)
    lower = int(max(0, (1.0 - sigma) * v))
    upper = int(min(255, (1.0 + sigma) * v))
    edged = cv2.Canny(image, lower, upper)
    return edged

G123_LUT = np.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0,
                     0, 1, 0, 1, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0,
                     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 1,
                     0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                     0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 1, 0, 1, 0, 0, 0, 1, 0, 1,
                     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0,
                     0, 0, 1, 0, 1, 0, 1, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0,
                     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                     0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1,
                     1, 0, 0, 1, 0, 0, 0, 1, 1, 0, 0, 1, 0, 0, 0, 1, 1, 0, 0,
                     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 1, 0, 0,
                     0, 1, 1, 0, 0, 1, 0, 0, 0], dtype=bool)

G123P_LUT = np.array([0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0,
                      0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0,
                      0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                      0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 1, 0, 1, 0, 0, 0, 0,
                      0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                      0, 0, 1, 0, 1, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                      0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0,
                      1, 0, 1, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0,
                      0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                      0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                      0, 0, 0, 1, 0, 1, 0, 1, 0, 1, 0, 0, 0, 0, 0, 1, 0, 1, 0,
                      0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 1,
                      0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                      0, 0, 0, 0, 0, 0, 0, 0, 0], dtype=bool)

def _generate_thin_luts():
    """generate LUTs for thinning algorithm (for reference)"""

    def nabe(n):
        return np.array([n >> i & 1 for i in range(0, 9)]).astype(bool)

    def G1(n):
        s = 0
        bits = nabe(n)
        for i in (0, 2, 4, 6):
            if not(bits[i]) and (bits[i + 1] or bits[(i + 2) % 8]):
                s += 1
        return s == 1

    g1_lut = np.array([G1(n) for n in range(256)])

    def G2(n):
        n1, n2 = 0, 0
        bits = nabe(n)
        for k in (1, 3, 5, 7):
            if bits[k] or bits[k - 1]:
                n1 += 1
            if bits[k] or bits[(k + 1) % 8]:
                n2 += 1
        return min(n1, n2) in [2, 3]

    g2_lut = np.array([G2(n) for n in range(256)])

    g12_lut = g1_lut & g2_lut

    def G3(n):
        bits = nabe(n)
        return not((bits[1] or bits[2] or not(bits[7])) and bits[0])

    def G3p(n):
        bits = nabe(n)
        return not((bits[5] or bits[6] or not(bits[3])) and bits[4])

    g3_lut = np.array([G3(n) for n in range(256)])
    g3p_lut = np.array([G3p(n) for n in range(256)])

    g123_lut = g12_lut & g3_lut
    g123p_lut = g12_lut & g3p_lut

    return bool2mask(g123_lut, value=1), bool2mask(g123p_lut, value=1)

if __name__=="__main__":
    image = cv2.imread(r"E:\PythonProject\pyzjrPyPi\pyzjr\test.png")
    mask, masked_image = create_rectmask(image, (50, 20), (180, 150))
    ulimg = up_low(image, (0, 0, 0), (255, 0, 255))
    approximg = approximate(image)
    ceilimg = ceilfloor(image)
    print(ceilimg.shape)
    # cv2.imshow("ss", mask)
    cv2.imshow("mask", ceilimg)
    cv2.waitKey(0)





















from .sampling import *
from .conv import *
from .atten import *
from .backbone import *
from .bricks import *
from .networks import get_clsnetwork
from ._utils import get_upsampling_weight, autopad, _make_divisible, channel_shuffle, convpool_outsize, make_layer
from .segmentation import *


from .check import *
from .mathfun import *
from .torch_np_unification import *
from .converter import (
    _ntuple, to_1tuple,
    to_2tuple, to_3tuple,
    to_4tuple, to_ntuple,
)

from .imtensor import (
    to_numpy,
    to_tensor,
    to_bchw,
    image_to_bchw,
    hwc2chw,
    chw2hwc,
    tensor_to_image,
    image_to_tensor,
    img2tensor,
    label2tensor
)

from .randfun import (
    rand,
    randint,
    randbool,
    randfluctuate,
    randlist,
    randintlist,
    randnormal,
    randshuffle,
    randstep,
    randstring,
    rand_choice_weighted
)